/*   This file is part of TownConstructor, which is a GameScript for OpenTTD
 *   Copyright (C) 2013  R2dical
 *
 *   This program is free software: you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation, either version 3 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
 
// Contributors:

// ToDo:
// use don't build industries rather than event and demolish

// Notes:
// Log levels:
//	static LVL_INFO = 1;           // main info. eg what it is doing
//	static LVL_SUB_DECISIONS = 2;  // sub decisions - eg. reasons for not doing certain things etc.
//	static LVL_DEBUG = 3;          // debug prints - debug prints during carrying out actions

// Imports
import("util.superlib", "SuperLib", 36);
Result 		<- SuperLib.Result;
Log 		<- SuperLib.Log;
Helper 		<- SuperLib.Helper;
ScoreList 	<- SuperLib.ScoreList;
Tile 		<- SuperLib.Tile;
Direction 	<- SuperLib.Direction;
Town 		<- SuperLib.Town;
Industry	<- SuperLib.Industry;
					
// Extend GS class
class TownConstructor extends GSController{
	MAP_SIZE_X = 1.0;
	MAP_SIZE_Y = 1.0;
	MAP_SCALE = 1.0;
	
	BUILD_TOWNS = false;										// Global build town check, set and checked in Init.
	BUILD_LIMIT = 0; 											// Set from settings, in class constructor and each refresh.(initial is max ind per map, subs is max per refresh)
	CONTINUE_GS = null;											// True if whole script must continue.
	INIT_PERFORMED = false;										// True if TownConstructor.Init has run.

	constructor(){
		INIT_PERFORMED = false;
		CONTINUE_GS = true;
		MAP_SIZE_X = GSMap.GetMapSizeX();
		MAP_SIZE_Y = GSMap.GetMapSizeY();
	}	
}
	
// Program start function
function TownConstructor::Start(){
	// Welcome msg
	Log.Info("Starting Industry Constructor!", Log.LVL_INFO);

	// Pause
	GSGame.Pause();
	// Call initialization function
	this.Init();
	// Unpause
	GSGame.Unpause();
	
	// Build town function
	if(BUILD_TOWNS == true){
		this.BuildTown();
	}
		
	// Display exit msg
	Log.Warning("The game script has ended, and will now crash.", Log.LVL_INFO);
}

// Initialization function
function TownConstructor::Init(){
	// Check GS continue
	if(CONTINUE_GS == false) return;
	
	// Display status msg
	Log.Info("+==============================+", Log.LVL_INFO);
	Log.Info("Initializing...", Log.LVL_INFO);
	
	// Check build towns
	// - Check script setting
	if(GSController.GetSetting("BUILD_TOWNS") == 1){
		// -- Valudate settings
		// -- Found Towns
		if(GSGameSettings.IsValid("found_town") == false)
			Log.Error("Found town setting could not be detected!", Log.LVL_INFO);
		// -- Road layouts
		else if(GSGameSettings.IsValid("town_layout") == false)
			Log.Error("Town road layout setting could not be detected!", Log.LVL_INFO);
		// -- City proportion
		else if(GSGameSettings.IsValid("larger_towns") == false)
			Log.Error("City proportion setting could not be detected!", Log.LVL_INFO);
		// -- City multiplier
		else if(GSGameSettings.IsValid("initial_city_size") == false)
			Log.Error("City size multiplier setting could not be detected!", Log.LVL_INFO);
		else
			this.BUILD_TOWNS = true;
	}
	// - Check number of towns on map
	if(this.BUILD_TOWNS == true && GSTown.GetTownCount() > 1)
	{
		BUILD_TOWNS = false;
		Log.Warning("Script set to build towns but more than 1 town on map! Script will disable this feature", Log.LVL_INFO);
	}
	
	// - Determine map multiplier, as float
	MAP_SCALE = (MAP_SIZE_X / 256.0) * (MAP_SIZE_Y / 256.0)
	
	// - Display status msg
	Log.Info(">Map size scale is: " + MAP_SCALE, Log.LVL_INFO);
	
	// Declare function status
	this.INIT_PERFORMED = true;
}

// Function call to build towns
function TownConstructor::BuildTown() {
	// Check GS continue, to exit
	if(CONTINUE_GS == false) return;
			
	// Display status msg
	Log.Info("+==============================+", Log.LVL_INFO);
	Log.Info("Building towns...", Log.LVL_INFO);

	// Check initiaized
	if (this.INIT_PERFORMED == false){
		// Display error msg
		Log.Error(">TownConstructor.BuildTown: Script uninitialized!", Log.LVL_INFO);
		return;
	}
			
	// Loop for number of towns modified by map size, not including initial 1
	local MAX_TOWNS = GSController.GetSetting("DENSITY_TOWN_TOTAL") * MAP_SCALE;
	local CITY_COUNT = 0;
	for(local i = 1; i < MAX_TOWNS; i++){	
		// Random build
		local BUILD_TRIES = 2000;
		local TILE_ID;
		local TOWN_SIZE;
		local IS_CITY = false;
		// Loop until correct tile
		while(BUILD_TRIES > 0){
			// Get a random tile
			TILE_ID = Tile.GetRandomTile();
			
			// Get a random size
			TOWN_SIZE = GSBase.RandRange(4);
			
			// Set city
			if(CITY_COUNT < GSGameSettings.GetValue("larger_towns")){
				IS_CITY = true;
				CITY_COUNT++;
			}
			
			local BUILT_TOWN = GSTown.FoundTown(TILE_ID, TOWN_SIZE, IS_CITY, GSGameSettings.GetValue("town_layout"), null);
			
			// Try build
			if(BUILT_TOWN == true){
				// Multiply size if city
				if(IS_CITY == true){
					// - Get town id
					local TOWN_ID = GSTile.GetTownAuthority(TILE_ID);
					// - Get house count
					local TOWN_HOUSES = GSTown.GetHouseCount(TOWN_ID);
					// Loop till size
					for(local j = 0; j < GSGameSettings.GetValue("initial_city_size"); j++)
						GSTown.ExpandTown(TOWN_ID, TOWN_HOUSES);
				}
				break;
			}	
			BUILD_TRIES--;
		}
		// Display error msg 
		if(BUILD_TRIES == 0)
			Log.Error(">TownConstructor.BuildTown: Failed to build town!", Log.LVL_INFO);	
	}
	
	// Display status msg
	local MAP_TOWNS = GSTown.GetTownCount();
	Log.Info("Built " + (MAP_TOWNS - CITY_COUNT) + " towns and " + CITY_COUNT + " cities, for a total of " +
		 MAP_TOWNS + " / " + MAX_TOWNS, Log.LVL_INFO);
}

// Function to check if tile is industry, returns true or false
function IsIndustry(TILE_ID) { return (GSIndustry.GetIndustryID(TILE_ID) != 65535); }

// Function to valuate town by dist from edge, returns the distance in integers
function GetTownDistFromEdge(TOWN_ID){
	return GSMap.DistanceFromEdge(GSTown.GetLocation(TOWN_ID));
}

// Function to check if a town is built on a desert tile, returns 1 if true or 0 if false
function IsTownDesert(TOWN_ID){
	// Get terrain type
	local TERRAIN_TYPE = GSTile.GetTerrainType(GSTown.GetLocation(TOWN_ID));
	
	// If desert return 1
	if(TERRAIN_TYPE == GSTile.TERRAIN_DESERT) return 1;
	
	// Else return 0
	else return 0;
	
}